import java.awt.geom.Rectangle2D;
import java.awt.font.FontRenderContext;
import java.awt.Font;
import java.awt.Color;
import javax.swing.JButton;
import java.awt.Graphics;
import javax.swing.ImageIcon;
import java.awt.Graphics2D;


/**
 * CS1322: Programming Assignment #6 - Spring 2003 - Phase 2
 *
 *  <PRE>
 *  Class PlayeLetterButton.java-Models a letter tile in Scrabble. Used as a
 *  button so that the user can just click on it to use it.
 *
 *  Revisions: 1.0 Apr. 17, 2003
 *               Created the PlayerLetterButton class
 *             1.1 Apr. 20, 2003
 *               Added java Docs
 *
 *  </PRE>
 *
 *  Collaboration statement:
 *  "I worked on the homework assignment alone, using only course materials."
 *
 *  @author <A HREF="mailto:gtg445h@mail.gatech.edu">Chris Wells</A>
 *  @version Version 1.0, Apr. 17, 2003
 */

public class PlayerLetterButton extends JButton implements ScrabbleConstants {

  /**
   * A constant to trigger debug events.
   */
  public final boolean bDEBUG = false;

  /**
   * The letter this button is representing.
   */
  private LetterModel letter;

  /**
     A conversion factor that determines where to place the value of
     letter relative to the position of the letter and scaled by the
     size of the button in the x direction.
  */
  private double valuePlacementXConvFactor;

  /**
     A conversion factor that determines where to place the value of
     letter relative to the position of the letter and scaled by the
     size of the button in the y direction.
  */
  private double valuePlacementYConvFactor;

  /**
     The font size that looks the best on a 800 by 600 pixel frame.  Scaled to
     fit current letter button size.
  */
  private final int DEFAULT_FONT_SIZE = 65;

  /**
     The height of the letter on a 800 x 600 pixel frame
  */
  private final double DEFAULT_HEIGHT = 100.0;

  /**
     Creates a new <code>PlayerLetterButton</code> with the entered letter
     as its representative letter (can be null).
  */
  public PlayerLetterButton(LetterModel letter) {
    this.letter = letter;
    valuePlacementXConvFactor = this.getWidth()/5;
    valuePlacementYConvFactor = this.getHeight()/9;
  }

  /**
     Overwrites the <code>paintComponent</code> to paint the letter if
     and value there is a valid letter, just a blank letter tile if it
     is a blank or just a tan background if the <code>letter</code> is
     <code>null</code>
  */
  protected void paintComponent(Graphics g) {
    Graphics2D g2 = (Graphics2D)g;
    if (letter == null) {
      g2.setColor(TAN);
      g2.fillRect(0,0,this.getWidth(), this.getHeight());
      this.setEnabled(false);
    }
    else {
      this.setEnabled(true);
      // a wood grain background picture
      ImageIcon letterBack = new ImageIcon("LetterTile.jpg");
      g2.drawImage(letterBack.getImage(),0,0,this.getWidth(),
		   this.getHeight(),this);
      // checks to see if the letter is not a blank.
      if (letter.getLetter() != 'b') {
	// font size of the letter. scaled by the height or width,
	// depenging on which one is smaller. of the button
	int letterSize = 0;
	if (this.getHeight() < this.getWidth()) { 
	  letterSize = (int) (DEFAULT_FONT_SIZE * 
			      (this.getHeight() / DEFAULT_HEIGHT));
	}
	else {
	  letterSize = (int) (DEFAULT_FONT_SIZE * 
			      (this.getWidth() / DEFAULT_HEIGHT));
	}

	if (bDEBUG)
	  System.out.println("LetterFontSize = " + letterSize);

	Font letterFont = new Font("SansSerif", Font.PLAIN, letterSize);
	g2.setFont(letterFont);
	g2.setColor(Color.white);
	// This next sequence of variables is used to perfectly center
	// the letter at all times
	String buttonLetter = String.valueOf(letter.getLetter());
	FontRenderContext context = g2.getFontRenderContext();
	Rectangle2D bounds = letterFont.getStringBounds(buttonLetter,context);
	double lAscent = -bounds.getY();
	double lDescent = bounds.getHeight() - lAscent;
	double lHeight = bounds.getHeight();
	double lWidth = bounds.getWidth();
	double lBase = ((this.getHeight() - lHeight) / 2) + lAscent;
	double lLeft = (this.getWidth() - lWidth) / 2;
	// the factor of how big the letter size should be to the
	// letter set as 3:1
	int letterToValueFontConv = 3;
	g2.drawString(buttonLetter, (float)lLeft, (float)lBase);
	// the font size of the value
	int valueSize = (int) (letterSize /letterToValueFontConv);

	if (bDEBUG)
	  System.out.println("ValueFontSize = " + valueSize);    

	Font valueFont = new Font("SansSerif", Font.PLAIN, valueSize);
	g2.setFont(valueFont);
	// places the value of the letter down and to the left of the letter
	String buttonValue = String.valueOf(letter.getValue());
	double vLeft = (double)(lLeft + lWidth + valuePlacementXConvFactor);
	double vBase = (double)(lBase + valuePlacementYConvFactor);
	g2.drawString(buttonValue, (float)vLeft, (float)vBase);
      }
    }
  }

  /**
   * Returns the letter model associated with this button.
   *
   * @return the letter model.
   */
  public LetterModel getLetter() {
    return letter;
  }

}// end of class PlayeLetterButton
