import java.util.StringTokenizer;
import java.util.Vector;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

/**
 * CS1322: Programming Assignment #5 - Spring 2003 - Phase 4
 *
 *  <PRE>
 *  Class ScrabbleFileReader.java- reads in the data file for Scrabble.
 *
 *  Revisions: 1.0 Apr. 04, 2003
 *               Created the ScrabbleFileReader class
 *
 *  </PRE>
 *
 *  Collaboration statement:
 *  "I worked on the homework assignment alone, using only course materials."
 *
 *  @author <A HREF="mailto:gtg445h@mail.gatech.edu">Chris Wells</A>
 *  @version Version 1.0, Apr. 04, 2003
 */

public class ScrabbleFileReader implements ScrabbleConstants {

  /**
   * A constant to trigger debug statements
   */

  public static final boolean bDEBUG = false;

  /**
   * Holds all of the <code>letters</code> pieces.
   *
   */
  private Vector letters;

  /**
   * Holds the representations of the <code>board</code> squares.
   *
   */
  private SquareModel[][] board;

  /**
   * Reads the data file for the scrabble game and intializes the 
   * <code>board</code> and letter pile.
   *
   */
  public ScrabbleFileReader() {
    this.letters = new Vector();
    this.board = new SquareModel[Y_LENGTH][X_LENGTH];
  }

  /**
   * Reads in the file ScrabbleDataFile.txt and assigns the letters and values
   * to the letters Vector and the board values and positions to the
   * board array.
   */
  public void readFile() {
    try{
      FileReader reader = new FileReader(INPUT_FILE);
      BufferedReader br = new BufferedReader(reader);
      final int LETTERS = 26; //number of letters in Scrabble (hopefully!)
      for (int i = 0; i < LETTERS; i++) {
	String currentRead = br.readLine();
	StringTokenizer current = new StringTokenizer(currentRead);
	if(bDEBUG)
	  System.out.println("Current line = " + currentRead);
	
	setLetters(current);
      }
      // adds the blank pieces in
      letters.add(new LetterModel('b', 0));
      letters.add(new LetterModel('b', 0));
      //gets the board model from the file
      for (int i = 0; i < Y_LENGTH; i++) {
	  String currentRead = br.readLine();
	  StringTokenizer current = new StringTokenizer(currentRead);
	  if(bDEBUG)
	    System.out.println("Current line = " + currentRead);
	  setSquares(current, i);
      }
      if (bDEBUG) {
	System.out.println("Letters -----------------");
	for(int i = 0; i < letters.size(); i++) {
	  System.out.println("Letter = " + 
			     ((LetterModel)letters.get(i)).getLetter() + 
			     ", Value = " + 
			     ((LetterModel)letters.get(i)).getValue());
	}
	for(int i = 0; i < Y_LENGTH; i++) {
	  String currentRow = "";
	  for(int j = 0; j < X_LENGTH; j++) {
	    currentRow = currentRow + "[" + board[i][j].getSquareType() + "] ";
	  }
	  System.out.println(currentRow);
	}
      }
    }
    catch(FileNotFoundException e) {
      System.err.println("File not found for some unknown reason." + '\n' +
			 "Exiting program cause of invalid data");
      System.exit(1);
    }
    catch(IOException e) {
      System.err.println("Exiting program cause of invalid data" + '\n' +
			 "Printing stack trace.");
      e.printStackTrace(new PrintWriter(System.err));
      System.exit(1);
    }
  }

  /**
   * Adds the correct number of a certain letter as contined in the
   * data file.
   *
   * @param letterInfo conatins the letter, number of letter to create
   * and value.
   */
  private void setLetters(StringTokenizer letterInfo) {
    try {
      String[] inputVals = new String[3]; // there are only 3 bits of data in the
                                          // data file for each letter entry
      int arrIndex = 0;
      while(letterInfo.hasMoreTokens()) {
	inputVals[arrIndex] = letterInfo.nextToken();
	arrIndex++;
      }
      int numOfLetters = Integer.parseInt(inputVals[1]);//value at 1 is the
                                                        //quanity of the letter
      for (int i = 0; i < numOfLetters; i++) {
	char letter = inputVals[0].charAt(0); //value at 0 is the letter
	int value = Integer.parseInt(inputVals[2]); //value at 2 is the value
	LetterModel currentLetter = new LetterModel(letter, value);
	letters.add(currentLetter);
      }
    }
    catch(NumberFormatException e) {
      System.err.println("Exiting program cause of invalid data" + '\n' +
			 "Printing stack trace.");
      e.printStackTrace(new PrintWriter(System.err));
      System.exit(1);
    }
  }

  /**
   * Sets each square according to the data file passed in.
   *
   * @param squareInfo contains a whole row of square values
   * @param yValue the y coordinate in the board of the row passed in
   */
  private void setSquares(StringTokenizer squareInfo, int yValue) {
    try {
      String[] inputVals = new String[X_LENGTH]; //there are X_LENGTH entries
                                                 // at each row in the data file
      int arrIndex = 0; // holds the x coord and array index
      int sqrType = 0;  // temp var to store each square type
      while(squareInfo.hasMoreTokens()) {
	inputVals[arrIndex] = squareInfo.nextToken();
	sqrType = Integer.parseInt(inputVals[arrIndex]);
	SquareModel currentSqr = new SquareModel(yValue, arrIndex, sqrType);
	board[yValue][arrIndex] = currentSqr;	
	arrIndex++;
      }
    }
    catch(NumberFormatException e) {
      System.err.println("Exiting program cause of invalid data" + '\n' +
			 "Printing stack trace.");
      e.printStackTrace(new PrintWriter(System.err));
      System.exit(1);
    }
  }

  /**
   * Gets the letters contained in Scrabble.
   *
   * @return the letters contained in Scrabble
   */
  public Vector getLetters() {
    return this.letters;
  }

  /**
   * Gets the board of Scrabble as a 2D array
   *
   * @return the board of Scrabble as a 2D array
   */
  public SquareModel[][] getBoard() {
    return this.board;
  }

}// end of class ScrabbleFileReader
