import java.awt.Graphics;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import java.util.Vector;

/**
 * CS1322: Programming Assignment #6 - Spring 2003 - Phase 4&5
 *
 *  <PRE>
 *  Class ScrabbleFrame.java-The frame which is the Scrabble game board.
 *  Also handles all user interaction for the game.
 *
 *  Revisions: 1.0 Apr. 20, 2003
 *               Created the ScrabbleFrame class
 *             1.1 Apr. 24, 2003
 *               Fixed undo to actually repaint
 *               Added the current letter frame
 *               Fixed putLetter to force horizantal and vertical word only
 *             1.2 Apr. 26, 2003
 *               Fixed undo error where the blank given back from undo has
 *               A letter on it.
 *             1.3 Apr. 26, 2003
 *               Fixed various bugs.
 *               Set letter would not allow words on edge sometimes.
 *               Setting blank letter in wrong place moved from setLetter to
 *               setSquare.
 *               Undo didn't reset wordYCoord and wordXCoord.
 *
 *  </PRE>
 *
 *  Collaboration statement:
 *  "I worked on the homework assignment alone, using only course materials."
 *
 *  @author <A HREF="mailto:gtg445h@mail.gatech.edu">Chris Wells</A>
 *  @version Version 1.2, Apr. 26, 2003
 */

public class ScrabbleFrame extends JFrame implements ScrabbleView, 
						     ScrabbleConstants,
						     ActionListener {
  /**
   * A constant to trigger debug statements.
   */
  public final boolean bDEBUG = false;

  /**
   * The currently selected letter.
   */
  private LetterModel current;

  /**
   * The frame which holds the current letter.
   */
  private JFrame currentFrame;

  /**
   * The model that does the mechanics of the game and saves states.
   */
  private Scrabble model;

  /**
   * The input returned from the start panel.
   */
  private String[] input;

  /**
   * The letters that have already been played in the squares they have
   * been placed in.
   */
  private Vector lettersPlaced;

  /**
   * The int parts of the constructor(numPlayers, startWordLength, start turn).
   */
  private int[] init1 = new int[3];

  /**
   * The boolean parts of the constructor(replaceBlanks, bingo)
   */
  private boolean[] init2 = new boolean[2];

  /**
   * The names of the players retrieved from <code>input[]</code>
   */
  private String[] names;

  /**
   * The number of blanks currently on the board.
   */
  private int blanksOnBoard;

  /**
   * The number of players in the game.  A number from 2-4.
   */
  private int numPlayers;

  /**
   * True if the variation to replace blanks on the board with the same
   * letter on their tray has been chosen, false otherwise.
   */
  private boolean replaceBlanks;

  /**
   * True if the variation to do Bingo rules has been chosen, false otherwise
   * (A Bingo is when all 7 letter on the players tray are played in one turn.)
   */
  private boolean bingo;

  /**
   * Holds the length the beginning word should be at least as long as.
   * Can be 2-7 but more than likely somewhere between 2-4.
   */
  private int startWordLength;

  /**
   * Determins weathere the first move has been made or not.
   * True if its the first move, false otherwise.
   */
  private boolean start;

  /**
   * True if the first word has not been played yet.
   */
  private boolean firstWord;

  /**
   * The scrabble board described in <code>SquareButton</code>s.
   */
  private SquareButton[][] board;

  /**
   * An array of letter buttons represting a players tray.
   */
  private PlayerLetterButton[] trayButtons;

  /**
   * An array of <code>JLabel</code>s to hold the scores of the players.
   */
  private JLabel[] scores;

  /**
   * An array of <code>JButton</code>s that holds the action buttons for
   * the game such as end turn, replace blank, challenge.
   */
  private JButton[] actionButtons;

  /**
   * The panel representing the "board".
   */
  private JPanel squarePanel;

  /**
   * Hold all the scores from <code>scores[]</code>.
   */
  private JPanel scorePanel;

  /**
   * Holds all the actionButtons.
   */
  private JPanel actionPanel;

  /**
   * A side picture that looks like the one on the scrabble board.
   */
  private JPanel sideBar;

  /**
   * The current players tray.
   */
  private JPanel tray;

  /**
   * Determines weather a player is replacing a blank piece or not.
   */
  private boolean replaceBlank;

  /**
   * Determines weather a player has requested to exchange letters or not
   */
  private boolean replaceLetters;

  /**
   * Determines whether trade letters is enabled or not.
   */
  private boolean canTrade;

  /**
   * The y coordinate and x coordinate of the placed word.
   */

  private int wordYCoord, wordXCoord;

  /**
   * Creates a new <code>ScrabbleFrame</code> and initializes all values.
   *
   * @param input a <code>String[]</code> in the form of <br>
   * index 0 - Number of Players (int) <br>
   *       1 - Starting Word Length (int) <br>
   *       2 - Starting turn (int) <br>
   *       3 - Bingo on/off (boolean) <br>
   *       4 - Replace Blanks on the Board on/off (boolean) <br>
   *     5-9 - The names of the player (String) could be 2 to 4 names
   * @param width the width of the frame
   * @param height the height of the frame
   */
  public ScrabbleFrame(String[] input, int width, int height) {
    this.input = input;
    start = true;
    firstWord = true;
    blanksOnBoard = 0;
    replaceLetters = false;
    replaceBlank = false;
    canTrade = true;
    lettersPlaced = new Vector();
    this.initialize();
    model = new ScrabbleModel(bingo, replaceBlanks, startWordLength,
			      numPlayers, Integer.parseInt(input[2]));
    this.initArrays();
    this.setSize(width, height);
    this.setLayout();
    this.initPanels();
    this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    this.update();
  }

  /**
   * A helper method to the constructor which intializes all the variation
   * variables and the names of the players.
   */
  private void initialize() {
    numPlayers = Integer.parseInt(input[0]);
    startWordLength = Integer.parseInt(input[1]);
    bingo = (Boolean.valueOf(input[3])).booleanValue();
    replaceBlanks = (Boolean.valueOf(input[4])).booleanValue();
    names = new String[numPlayers];
    for (int i = 0; i < numPlayers; i++) {
      // i + 5 because the names start at input index 5;
      names[i] = input[i + 5]; 
    }
    if (bDEBUG){
      System.out.println("# of players = " + numPlayers + " Start word length"
			 + startWordLength + " Bingo? " + bingo + '\n' + 
			 "Replace blanks? " + replaceBlanks);
      for (int i = 0; i < numPlayers; i++) {
	System.out.println("Player " + i + " name is - " + names[i]);	  
      }       
    }
    wordXCoord = -1;
    wordYCoord = -1;
  }

  /**
   * Initializes all arrays to their initial values other than input.
   */
  private void initArrays() {
    scores = new JLabel[numPlayers];
    for(int i = 0; i < numPlayers; i++) {
      scores[i] = new JLabel(names[i] + ": " 
			     + model.getScore(i));
      scores[i].setBackground(TAN);
      scores[i].setForeground(Color.black);
      scores[i].setToolTipText(names[i] + " score which currently is " +
			       model.getScore(i));
    }

    board = new SquareButton[Y_LENGTH][X_LENGTH];
    for (int i = 0; i < Y_LENGTH; i++) {
      SquareModel square = null;
      for (int j = 0; j < X_LENGTH; j++) {
	square = model.getSquare(i,j);
	board[i][j] = new SquareButton(square);
	board[i][j].addActionListener(this);
      }
    }

    actionButtons = new JButton[6];
    actionButtons[0] = new JButton("Trade in Letters");
    actionButtons[0].setToolTipText("Press to exchange letters.");
    actionButtons[0].setActionCommand("Trade");
    actionButtons[1] = new JButton("Replace Blank");
    actionButtons[1].setToolTipText("Press to replace a blank on the board.");
    actionButtons[1].setEnabled(false);
    actionButtons[1].setActionCommand("Blank");
    actionButtons[2] = new JButton("Undo placement");
    actionButtons[2].setActionCommand("Undo");
    actionButtons[3] = new JButton(names[model.getTurn()] +
				   " End Turn");
    actionButtons[3].setActionCommand("EndTurn");
    // extra buttons for other options that come up during a turn after certain
    // buttons were pressed.
    actionButtons[4] = new JButton();
    actionButtons[5] = new JButton();
    actionButtons[4].setEnabled(false);
    actionButtons[5].setEnabled(false);
    for (int i = 0; i < actionButtons.length; i++) {
      actionButtons[i].setBackground(LIGHT_TAN);
      actionButtons[i].setForeground(Color.black);
      actionButtons[i].addActionListener(this);
    }

    trayButtons = new PlayerLetterButton[TRAY_SIZE];
  }

  /**
   * Makes the panels compatiable with a <code>GridBagLayout</code> so that
   * the layout and sizes of the panels are displayes right.
   */
  private JPanel makePanel(GridBagLayout gridBag,
			   GridBagConstraints c) {
    JPanel panel = new JPanel();
    gridBag.setConstraints(panel, c);
    return panel;
  }

  /**
   * Sets up the layout for the frame.  Uses a <code>GridBagLayout</code>
   * to make the proportions always correct.
   */
  private void setLayout() {
    Graphics g = this.getGraphics();
    // setting up a grid bag layout to ensure proper aligment and size of
    // panels
    GridBagLayout gridBag = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();
    this.getContentPane().setLayout(gridBag);
    
    c.fill = GridBagConstraints.BOTH;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.gridheight = 3; 
    c.weighty = .5;
    scorePanel = makePanel(gridBag, c);
    this.getContentPane().add(scorePanel);
    
    c.gridwidth = 3; 
    c.gridheight = GridBagConstraints.RELATIVE; 
    c.weighty = 12.0;
    c.weightx = 3.0;
    sideBar = new SideBar(gridBag, c, "ScrabbleSide.jpg");
    this.getContentPane().add(sideBar);
    
    c.gridheight = GridBagConstraints.RELATIVE; 
    c.weightx = 10.0;
    squarePanel = makePanel(gridBag, c);
    this.getContentPane().add(squarePanel);
    
    c.gridwidth = GridBagConstraints.REMAINDER; 
    c.weightx = .5;
    actionPanel = makePanel(gridBag, c);
    this.getContentPane().add(actionPanel);
    
    c.gridwidth = GridBagConstraints.REMAINDER; 
    c.gridheight = GridBagConstraints.REMAINDER;
    c.weighty = 2.0;
    tray = makePanel(gridBag, c);
    this.getContentPane().add(tray);
  }

  /**
   * Initializes all panel variables.
   */
  private void initPanels() {
    squarePanel.setLayout(new GridLayout(Y_LENGTH, X_LENGTH));
    for (int i = 0; i < Y_LENGTH; i++) {
      for (int j = 0; j < Y_LENGTH; j++) {
	squarePanel.add(board[i][j]);	
      }
    }

    scorePanel.setLayout(new GridLayout(1, numPlayers));
    for (int i = 0; i < scores.length; i++) {
      scorePanel.add(scores[i]);
    }

    actionPanel.setLayout(new GridLayout(actionButtons.length, 1));
    for (int i = 0; i < actionButtons.length; i++) {
      actionPanel.add(actionButtons[i]);
    }

    tray.setLayout(new GridLayout(1, TRAY_SIZE + 1));
    tray.add(new JLabel(names[model.getTurn()] + "'s Tray"));
    Vector playerLetters = model.getPlayerLetters(model.getTurn());
    LetterModel current = null;
    for (int i = 0; i < TRAY_SIZE; i++) {
      current = (LetterModel)playerLetters.get(i);
      trayButtons[i] = new PlayerLetterButton(current);
      trayButtons[i].addActionListener(this);
      tray.add(trayButtons[i]);
    }

  }

  /**
   * Needed by the interface <code>ActionListener</code> checks to see if
   * the start or quit buttons have been pressed or one of the choices has
   * changed.
   *
   * @param event the event thrown by the button that was pressed
   */
  public void actionPerformed(ActionEvent event) {
    if (event.getActionCommand().equals("Trade") && canTrade) {
      this.trade();
    }
    else if (event.getActionCommand().equals("Blank")) {
      this.replBlank();
    }
    else if (event.getActionCommand().equals("Undo")) {
      this.undo();
    }
    else if (event.getActionCommand().equals("EndTurn")) {
      this.endTurn();
    }
    else if (event.getActionCommand().equals("UndoLast")) {
      this.undoLast();
    }
    else if (event.getActionCommand().equals("UndoAll")) {
      this.undoAll();
    }
    else if (event.getActionCommand().equals("Challenge")) {
      this.challenge();
    }
    else if (event.getActionCommand().equals("NoChallenge") ||
	     event.getActionCommand().equals("LegalWord")) {
      this.endTurnLegal();
    }
    else if (event.getActionCommand().equals("IllegalWord")) {
      this.illegalWord();
    }
    else {
      if (bDEBUG)
	System.out.println("event source" + event.getSource().getClass());
      
      if(event.getSource() instanceof PlayerLetterButton) {
	if (bDEBUG)
	  System.out.println("going to set letter");

	current = null;
	this.setLetter((PlayerLetterButton)event.getSource());
      }
      else if((event.getSource() instanceof SquareButton) 
	      && !replaceLetters && !replaceBlank && current != null) {
	if (bDEBUG)
	  System.out.println("going to put letter");
	      
	this.putLetter((SquareButton)event.getSource());
      }
    }
  }

  /**
   * Acitvates trading in letters.
   */
  private void trade() {
    replaceLetters = true;
    actionButtons[0].setEnabled(false);
    actionButtons[1].setEnabled(false);
    model.setCurrent(new LetterModel('^', 0));
    JOptionPane.showMessageDialog(this.getContentPane(), 
				  "Click on the letters you want to"
				  + " replace then hit End Turn",
				  "Trade Letters Info",
				  JOptionPane.INFORMATION_MESSAGE);
  }

  /**
   * Sets up the ability to trade in you letter for a blank on the board.
   */
  private void replBlank() {
    replaceBlank = true;
    actionButtons[0].setEnabled(false);
    actionButtons[1].setEnabled(false);
    model.setCurrent(new LetterModel('#', 0));
    JOptionPane.showMessageDialog(this.getContentPane(),
				  "Click on a letter to try to replace a blank"+
				  " if there is one on the board and it matches"
				  +  " your letter then you get it",
				  "Replace Blanks Info",
				  JOptionPane.INFORMATION_MESSAGE);
  }

  /**
   * Creates the menu for undo.
   */
  private void undo() {
    current = null;
    if(currentFrame != null){
      currentFrame.dispose();
    }
    actionButtons[4].setText("Undo Last");
    actionButtons[4].setActionCommand("UndoLast");
    actionButtons[4].setEnabled(true);
    actionButtons[5].setText("Undo All");
    actionButtons[5].setActionCommand("UndoAll");
    actionButtons[5].setEnabled(true);
    actionButtons[0].setEnabled(false);
    actionButtons[1].setEnabled(false);
    actionButtons[2].setEnabled(false);
    JOptionPane.showMessageDialog(this.getContentPane(),
				  "Choose undo last to undo the" + 
				  " last letter placed and reset all pressed" +
				  " buttons this turn (still your turn)",
				  "Undo Info",
				  JOptionPane.INFORMATION_MESSAGE);

  }

  /**
   * Undoes your last letter placement.
   */
  private void undoLast() {
    model.setCurrent(new LetterModel('!', 0));
    if (replaceLetters && (lettersPlaced.size() != 0)) {
      // already a LetterModel
      Object last = lettersPlaced.get(lettersPlaced.size() - 1);
      model.getPlayerLetters(model.getTurn()).add(last);
      lettersPlaced.remove(last);
      this.update();
    }
    else if (lettersPlaced.size() != 0) {
      SquareModel last = 
	(SquareModel)lettersPlaced.get(lettersPlaced.size() - 1);
      int xCoord = last.getXCoord();
      int yCoord = last.getYCoord();

      Object lastPlaced = last.getLetter();
      if (((LetterModel)lastPlaced).getValue() == 0)
	model.getPlayerLetters(model.getTurn()).add(new LetterModel('b',0));
      else
	model.getPlayerLetters(model.getTurn()).add(lastPlaced);
      last.setLetter(null);
      board[yCoord][xCoord] = new SquareButton(last);
      lettersPlaced.remove(last);
      squarePanel.paintImmediately(0, 0, squarePanel.getWidth(),
				   squarePanel.getHeight());
      this.update();
    }
    this.resetExtraButtons();
    actionButtons[0].setEnabled(true);
    replaceBlank = false;
    if (lettersPlaced.size() == 0) {
      wordXCoord = -1;
      wordYCoord = -1;
      canTrade = true;
    }

    if (model.getSquare(Y_LENGTH/2, X_LENGTH/2).getLetter() == null){
      start = true;
      firstWord = true;
    }

    this.update();
  }

  /**
   * Undoes all letter placements made during this turn.
   */
  private void undoAll() {
    int size = lettersPlaced.size();
    for (int i = 0; i <= size; i++) {
      this.undoLast();
    }
    this.update();
  }

  /**
   * Sets up the challenge word menu.
   */
  private void challenge() {
    actionButtons[4].setText("Legal Word");
    actionButtons[4].setActionCommand("LegalWord");
    actionButtons[4].setEnabled(true);
    actionButtons[5].setText("Illegal Word");
    actionButtons[5].setActionCommand("IllegalWord");
    actionButtons[5].setEnabled(true);
    actionButtons[0].setEnabled(false);
    actionButtons[1].setEnabled(false);
    actionButtons[2].setEnabled(false);
    JOptionPane.showMessageDialog(this.getContentPane(),
				  "Clicking Illegal word mean that the players"
				  + " agree that the word in question is not" +
				  " a legal word and the players letters are" +
				  " returned to him but he loses his turn",
				  "Challenge!",
				  JOptionPane.INFORMATION_MESSAGE);
  }

  /**
   * detemines what to do when end turn is pressed.
   */
  private void endTurn() {
    wordXCoord = -1;
    wordYCoord = -1;
    if(currentFrame != null){
      currentFrame.dispose();
    }
    if (!replaceLetters && (lettersPlaced.size() != 0)) {
      if (!firstWord || lettersPlaced.size() >= startWordLength) {
	actionButtons[4].setText("Challenge Word");
	actionButtons[4].setActionCommand("Challenge");
	actionButtons[4].setEnabled(true);
	actionButtons[5].setText("No Challenge");
	actionButtons[5].setActionCommand("NoChallenge");
	actionButtons[5].setEnabled(true);
	actionButtons[0].setEnabled(false);
	actionButtons[1].setEnabled(false);
	actionButtons[2].setEnabled(false);
	JOptionPane.showMessageDialog(this.getContentPane(),
				      "Anyone can click the challenge button to" +
				      " challenge the words the person just made" +
				      " if there is no challenge the words stay" +
				      " and play continues",
				      "Challenge?",
				      JOptionPane.INFORMATION_MESSAGE);
	firstWord = false;
      }
      else {
	JOptionPane.showMessageDialog(this.getContentPane(),
				      "Starting word is not long enough." +
				      "It should be " + startWordLength + 
				      " letters long, it is only " +
				      lettersPlaced.size() + " letters long.",
				      "Illegal Move",
				      JOptionPane.WARNING_MESSAGE);
	return;
      }
    }
    else {
      this.endTurnLegal();
    }
  }

  /**
   * Ends turn when a legal word has been played.
   */
  private void endTurnLegal() {
    lettersPlaced = new Vector();
    replaceLetters = false;
    replaceBlank = false;
    canTrade = true;
    this.resetExtraButtons();
    update();
    model.endTurn();
    JOptionPane.showMessageDialog(this.getContentPane(),
				  "It is now " + names[model.getTurn()] +
				  "'s turn",
				  "Turn Change",
				  JOptionPane.INFORMATION_MESSAGE);
    update();
  }

  /**
   * Special method for ending turn on an illegal word.
   */
  private void illegalWord() {
    this.undoAll();
    this.endTurnLegal();
  }

  /**
   * Sets the current letter or replaces blanks or letters.
   */
  private void setLetter(PlayerLetterButton button) {
    if(bDEBUG)
      System.out.println("Entering set Letter");
    
    if (replaceBlank) {
      if(bDEBUG)
	System.out.println("Replacing Blank");
      
      boolean alreadyBlank = false;
      int last = model.getPlayerLetters(model.getTurn()).size() - 1;
      LetterModel letter = (LetterModel)
	model.getPlayerLetters(model.getTurn()).get(last);
      if (letter.getLetter() == 'b') 
	alreadyBlank = true;

      model.setCurrent(button.getLetter());

      // if the last letter in their tray is a blank then it means that
      // the method worked.
      if (letter.getLetter() == 'b' && !alreadyBlank) 
	blanksOnBoard = blanksOnBoard - 1;

      replaceBlank = false;
      this.update();
    }

    else if (replaceLetters) {
      if(bDEBUG)
	System.out.println("Setting letter to replace");
      
      lettersPlaced.add(button.getLetter());
      model.setCurrent(button.getLetter());
      this.update();
    }
    else {
      if(bDEBUG)
	System.out.println("Setting current ? ");
      current = button.getLetter();      
      model.setCurrent(button.getLetter());
      this.showCurrent();
    }
  }

  /**
   * Puts a letter on a square.
   */
  private void putLetter(SquareButton button) {
    if(bDEBUG)
      System.out.println("letter placed size = " + lettersPlaced.size());
    
    SquareModel square = button.getSquare();
    int xCoord = square.getXCoord();
    int yCoord = square.getYCoord();
    String orientation = "";
    // new word reset coordinates
    
    // gets orientation of word by seeing if the word is horiz or vert by checking
    // the x coordinates of the first two letters
    if (lettersPlaced.size() > 1) {
      if (((SquareModel)(lettersPlaced.get(0))).getXCoord() ==
	  ((SquareModel)(lettersPlaced.get(1))).getXCoord())
	orientation = "vert";
      else
	orientation = "horiz";
    }
    
    // abstraction booleans for if the letter is on one of the edges
    boolean rightEdge = xCoord == (X_LENGTH - 1);
    boolean leftEdge = xCoord == 0;
    boolean topEdge = yCoord == (Y_LENGTH - 1);
    boolean bottomEdge = yCoord == 0;
    
    // abstraction booleans for if there is a letter to the left, right, up
    // and down of the place letter if its on that edge it doesnt need to 
    // check
    boolean letterLeft = false;
    boolean letterRight = false;
    boolean letterUp = false;
    boolean letterDown = false;
    if (!leftEdge)
      letterLeft = model.getSquare(yCoord, xCoord - 1).getLetter() != null;
    if (!rightEdge)
      letterRight = model.getSquare(yCoord, xCoord + 1).getLetter() != null;
    if (!topEdge)
      letterUp = model.getSquare(yCoord + 1, xCoord).getLetter() != null;
    if (!bottomEdge)
      letterDown = model.getSquare(yCoord - 1, xCoord).getLetter() != null;
    
    if(bDEBUG)
      System.out.println("right edge, letter " + rightEdge + ", " + letterRight +
			 " left edge, letter " + leftEdge + ", " + letterLeft +
			 "\ntop edge, letter " + topEdge + ", " + letterUp +
			 " bottom edge, letter " + bottomEdge + ", " + letterDown
			 + "\n word (y, x) -(" + wordYCoord + ", " +
			 wordXCoord + " )");
    
    if (start && square.getSquareType() == START) {
      this.setSquare(square, xCoord, yCoord);
      start = false;
      return;
    }
    else if (orientation.equals("") && start == false) {
      // makes sure that one of the squares around this one has a letter
      // or the letter is on the edge and one of the ones around it still
      // has a letter and it is the first or second letter placed


      if(((rightEdge && letterLeft) || (leftEdge && letterRight) ||
	  (topEdge && letterDown) || (bottomEdge && letterUp)) &&
	 (xCoord == wordXCoord || yCoord == wordYCoord 
	  || wordXCoord == -1)) {	
	this.setSquare(square, xCoord, yCoord);
	return;
      }
      
      if (((topEdge && (letterLeft || letterRight)) || 
	   (bottomEdge && (letterLeft || letterRight)) || 
	   (rightEdge && (letterUp || letterDown)) || 
	   (leftEdge && (letterUp || letterDown))) &&
	  (xCoord == wordXCoord || yCoord == wordYCoord 
	   || wordXCoord == -1)) {
	this.setSquare(square, xCoord, yCoord);
	return;
      }
      
      else if (topEdge || bottomEdge || rightEdge || leftEdge) {
	// "breaks" if
      }
      
      else if ((letterRight || letterLeft || letterUp || letterDown) &&
	       (xCoord == wordXCoord || yCoord == wordYCoord
		|| wordXCoord == -1)) {
	this.setSquare(square, xCoord, yCoord);
	return;
      }
      
    }
    // checks orientation and makes sure letter is near another letter and
    // in the appropriate orientation

    else if (orientation.equals("horiz")) {
      if ((rightEdge && letterLeft) || (leftEdge && letterRight) &&
	  (yCoord == wordYCoord)) {
	this.setSquare(square, xCoord, yCoord);
	return;
      }
      
      else if (rightEdge || leftEdge) {
	// "breaks" if
      }

      else if ((letterRight || letterLeft) && (yCoord == wordYCoord)) {
	this.setSquare(square, xCoord, yCoord);
	return;
      }
    }

    else if (orientation.equals("vert")) {
      if (((topEdge && letterDown) || (bottomEdge && letterUp)) && 
	  (xCoord == wordXCoord)){
	this.setSquare(square, xCoord, yCoord);
	return;
      }
      
      else if (topEdge || bottomEdge) {
	// "breaks" if	
      }

      else if ((letterUp || letterDown) && (xCoord == wordXCoord)){
	this.setSquare(square, xCoord, yCoord);
	return;
      }
    }
    this.invalidSquare();
  }

  /**
   * This sets the current letter to the square.
   *
   * @param square the square to change
   * @param xCoord the x Coordinate of the square
   * @param yCoord the y Coordinate of the square
   */
  private void setSquare(SquareModel square, int xCoord, int yCoord) {
    if (current.getValue() == 0) { 
      if(bDEBUG)
	System.out.println("Setting Blanks on Board ? ");
      
      blanksOnBoard = blanksOnBoard + 1;
      boolean done = false;
      String input = "";
	while (!done) {
	  input = JOptionPane.showInputDialog(this.getContentPane(),
					      "Choose a letter for the blank" +
					      " to represent (A-Z only, upper"
					      + "-case only)",
					      "Choose Blank",
					      JOptionPane.QUESTION_MESSAGE);
	  if (input == null)
	  JOptionPane.showMessageDialog(this.getContentPane(),
					"Choose a letter, PLEASE",
					"Choose a valid letter",
					JOptionPane.WARNING_MESSAGE);
	  else
	  done = true;
	}
      char in = input.charAt(0);
      model.getPlayerLetters(model.getTurn()).remove(current);
      LetterModel blank = new LetterModel(in, 0);
      model.getPlayerLetters(model.getTurn()).add(blank);
      current = blank;
      model.setCurrent(blank);	   
    }
    model.squareSelect(square);
    lettersPlaced.add(square);
    board[yCoord][xCoord] = new SquareButton(square);
    board[yCoord][xCoord].repaint();
    // new word reset coordinates
    if (lettersPlaced.size() == 1) {
      wordXCoord = xCoord;
      wordYCoord = yCoord;
    }
    else if (lettersPlaced.size() == 2) {
      if (xCoord == wordXCoord)
	wordXCoord = xCoord;
      else if (yCoord == wordYCoord)
	wordYCoord = yCoord;
    }
    current = null;
    canTrade = false;
    if(currentFrame != null){
      currentFrame.dispose();
    }
    this.update();
  }

  private void invalidSquare() {
    if(bDEBUG)
      System.out.println("Calling invalid square");
     
    JOptionPane.showMessageDialog(this.getContentPane(),
				  "Either you tried to put a letter not next" +
				  " to another one or you tried to put a" +
				  " letter somewhere other than the start" +
				  " square on the first turn.  Pick another" +
				  " square.",
				  "Illegal Move",
				  JOptionPane.WARNING_MESSAGE);
  }

  /**
   * Shows the current letter frame.
   */
  private void showCurrent() {
    if(bDEBUG)
      System.out.println("Adding label and letter");
    if(currentFrame != null){
      currentFrame.dispose();
    }
    currentFrame = new JFrame("Letter");    
    currentFrame.setSize(80, 160);
    currentFrame.setLocation(this.getWidth() - currentFrame.getWidth(),
			     this.getHeight()- currentFrame.getHeight());
    currentFrame.setResizable(false);
    currentFrame.getContentPane().setLayout(new GridLayout(2, 1));
    currentFrame.setVisible(true);
    JLabel label = new JLabel("Current letter is ");
    PlayerLetterButton letter = new PlayerLetterButton(current);
    currentFrame.getContentPane().add(label);
    currentFrame.getContentPane().add(letter);
    label.repaint();
    letter.repaint();
    currentFrame.validate();
  }

  /**
   * Resets the extra two functionality buttons to nothing.
   */
  private void resetExtraButtons() {
    actionButtons[4].setText("");
    actionButtons[4].setActionCommand("");
    actionButtons[4].setEnabled(false);
    actionButtons[5].setText("");
    actionButtons[5].setActionCommand("");
    actionButtons[5].setEnabled(false);
  }

  /**
   * Updates all parts of the gameboard.
   */
  public void update() {
    if (model.gameIsOver())
      this.endGame();
    if ((blanksOnBoard > 0) && replaceBlanks)
      actionButtons[1].setEnabled(true);
    else
      actionButtons[1].setEnabled(false);

    if (canTrade)
      actionButtons[0].setEnabled(true);
    else
      actionButtons[0].setEnabled(false);

    if (lettersPlaced.size() > 0)
      actionButtons[2].setEnabled(true);
    else
      actionButtons[2].setEnabled(false);


    tray.removeAll();
    int currentTurn = model.getTurn();
    Vector currentTray = model.getPlayerLetters(currentTurn);
    tray.add(new JLabel(names[model.getTurn()] + "'s Tray"));
    for (int i = 0; i < TRAY_SIZE; i++) {
      if (i >= currentTray.size()) {
	trayButtons[i] = new PlayerLetterButton(null);
      }
      else {
	trayButtons[i] = new PlayerLetterButton((LetterModel)currentTray.get(i));
      }
      trayButtons[i].addActionListener(this);
      tray.add(trayButtons[i]);
    }
    actionButtons[3].setText(names[model.getTurn()] +
			     " End Turn");
    tray.validate();
    actionPanel.validate();
    squarePanel.validate();
    scorePanel.removeAll();
    for(int i = 0; i < numPlayers; i++) {
      scores[i] = new JLabel(names[i] + ": " 
			     + model.getScore(i));
      scores[i].setBackground(TAN);
      scores[i].setForeground(Color.black);
      scores[i].setToolTipText(names[i] + " score which currently is " +
			       model.getScore(i));
    }
    for (int i = 0; i < scores.length; i++) {
      scorePanel.add(scores[i]);
    }
    scorePanel.validate();
    this.validate();

  }

  /**
   * Ends current game and asks to play another.
   */
  private void endGame() {
    lettersPlaced = new Vector();
    canTrade = false;
    replaceBlanks = false;
    for (int i = 0; i < Y_LENGTH; i++) {
      for (int j = 0; j < Y_LENGTH; j++) {
	board[i][j].setEnabled(false);	
      }
    }

    for (int i = 0; i < TRAY_SIZE;i++) {
      trayButtons[i].setEnabled(false);
    }
    for (int i = 0; i < actionButtons.length;i++) {
      actionButtons[i].setEnabled(false);
    }

    tray.removeAll();
    tray.validate();
    actionPanel.validate();
    squarePanel.validate();
    scorePanel.removeAll();
    for(int i = 0; i < numPlayers; i++) {
      scores[i] = new JLabel(names[i] + ": " 
			     + model.getScore(i));
      scores[i].setBackground(TAN);
      scores[i].setForeground(Color.black);
      scores[i].setToolTipText(names[i] + " score which currently is " +
			       model.getScore(i));
    }
    for (int i = 0; i < scores.length; i++) {
      scorePanel.add(scores[i]);
    }
    scorePanel.validate();

    this.validate();

    int[] scores = new int[numPlayers];
    for (int i = 0; i < numPlayers;i++) {
      scores[i] = model.getScore(i);
    }
    int[] winner = new int[2];
    int[] ranks = new int[numPlayers];
    for (int i = 0; i < numPlayers; i++) {
      for (int j = 0; j < numPlayers; j++) {
	if (i != j && scores[i] >= scores[j])
	  ranks[i] = ranks[i] + 1;
      }  
    }
    winner[1] = 0;
    for (int i = 0; i < numPlayers; i++) {
      if (ranks[i] == (numPlayers - 1) && winner[0] == 0) {
	winner[0] = i;
      }
      else if(ranks[i] == (numPlayers - 1)) {
	winner[1] = i;
      }
    }

    String message = "The winner is " + names[winner[0]];
    if (winner[1] > 0)
      message = message + " and " + names[winner[1]];
    message= message + "\nWho had " + scores[winner[0]] + " points.";
    message = message + "\nWould you like to play again?";
    int again = JOptionPane.showConfirmDialog(this.getContentPane(),
					      message,
					      "Game Over",
					      JOptionPane.YES_NO_OPTION,
					      JOptionPane.QUESTION_MESSAGE);
    if (again == JOptionPane.NO_OPTION){
      this.dispose();
      thanks();
    }
    else {
      this.dispose();
      ScrabbleInit game = new ScrabbleInit();
      game.setVisible(true);
      game.validate();
    }
  }

  /**
   * My thanks for playing frame.
   */
  private void thanks() {
    JFrame thanks = new JFrame();
    thanks.setSize(300, 300);
    thanks.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    thanks.setVisible(true);
    thanks.setTitle("Thanks for Playing");
    thanks.getContentPane().add(new MySigPanel("Thanks for playing.  Coded by"));
    thanks.validate();
  }

}// end of class ScrabbleFrame
