import java.awt.GridLayout;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.ButtonGroup;
import javax.swing.JRadioButton;
import javax.swing.JOptionPane;
import javax.swing.JFrame;
import javax.swing.JWindow;
import javax.swing.JTextField;
import javax.swing.border.BevelBorder;
import javax.swing.border.TitledBorder;
import java.awt.Font;
import java.awt.Component;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.Vector;
import java.util.StringTokenizer;

/**
 * CS1322: Programming Assignment #6 - Spring 2003
 *
 *  <PRE>
 *  Class ScrabbleInit.java-Asks for user to select variations, num of players
 *  etc.
 *
 *  Revisions: 1.0 Apr. 21, 2003
 *               Created the ScrabbleInit class
 *             1.1 Apr. 24, 2003
 *               Added screen size
 *
 *  </PRE>
 *
 *  Collaboration statement:
 *  "I worked on the homework assignment alone, using only course materials."
 *
 *  @author <A HREF="mailto:gtg445h@mail.gatech.edu">Chris Wells</A>
 *  @version Version 1.1, Apr. 24, 2003
 */

public class ScrabbleInit extends JFrame implements ActionListener{

  /**
   * A constant to trigger debug statements.
   */
  public final boolean bDEBUG = false;

  /**
   * Holds inputs for game as a <code>String[]</code>.  Set up as
   * index 0 - Number of Players (int) <br>
   *       1 - Starting Word Length (int) <br>
   *       2 - Starting turn (int) <br>
   *       3 - Bingo on/off (boolean) <br>
   *       4 - Replace Blanks on the Board on/off (boolean) <br>
   *     5-9 - The names of the player (String) could be 2 to 4 names
   */
  private String[] input;

  /**
   * Hold the number of players as an int.  Used often in this class.
   */
  private int numPlayers;

  /**
   * Holds the three choices for number of players (2, 3, 4).
   */
  private JRadioButton[] players;

  /**
   * Holds the six choices for start word length (2, 3, 4, 5, 6, 7).
   */
  private JRadioButton[] wordLengths;

  /**
   * Holds the yes and no buttons to the replace blanks choice.
   */
  private JRadioButton[] replaceBlanks;

  /**
   * Holds the screen size buttons (640x480, 800x600, 1024x768, 1280x1024,
   * 1600x1200)
   */
  private JRadioButton[] screenSize;

  /**
   * Constant to hold screen size lengths
   */
  private String[] SCREEN_SIZE = {"640x480", "800x600", "1040x768", "1280x1040",
				  "1600x1200"};

  /**
   * Holds the yes and no buttons to the bingo choice.
   */
  private JRadioButton[] bingoButtons;

  /**
   * Holds the text fileds for the names of the players
   */
  private JTextField[] names;

  /**
   * The panels to hold the different choices.
   */
  private JPanel[] choicePanels = new JPanel[7];

  /**
   * The window to show the letters the players drew.
   */
  private JWindow letters;

  /**
   * Holds the start size of the game window. width in index 0 height in 1.
   */
  private int[] gameSize;

  /**
   * Construct a new <code>ScrabbleInit</code> frame.  Initializes all
   * instance fields.
   */
  public ScrabbleInit() {
    this.setTitle("Scrabble Start");
    input = new String[9]; // 9 is max number of inputs
    this.getContentPane().setLayout(new GridLayout(1, 2));
    players = new JRadioButton[3];
    replaceBlanks = new JRadioButton[2];
    bingoButtons = new JRadioButton[2];
    wordLengths = new JRadioButton[6];
    screenSize = new JRadioButton[SCREEN_SIZE.length];
    names = new JTextField[4];
    gameSize = new int[2];
    input[0] = "2"; //sets players to a default of 2
    numPlayers = Integer.parseInt(input[0]);
    addComponents();
    this.setSize(500,400);
    this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    this.validate();
  }

  /**
   * Needed by the interface <code>ActionListener</code> checks to see if
   * the start or quit buttons have been pressed or one of the choices has
   * changed.
   *
   * @param event the event thrown by the button that was pressed
   */
  public void actionPerformed(ActionEvent event) {
    if(bDEBUG)
      System.out.println("Something happened " + event.getActionCommand());
    if(event.getActionCommand().equals("Quit")) {
      this.dispose();
      thanks();
    }
    else if(event.getActionCommand().equals("Done")) {
      this.setInput();
      input[2] = String.valueOf(drawTurn());
      JButton start = new JButton("Start Game");
      start.setActionCommand("Start");
      start.addActionListener(this);
      choicePanels[3].add(start);
      choicePanels[3].repaint();
      for (int i = 0; i < choicePanels.length; i++) {
	Component[] components = choicePanels[i].getComponents();
	for (int j = 0; j < components.length; j++) {
	  components[j].setEnabled(false);
	}
      }
      start.setEnabled(true);
      choicePanels[3].getComponent(1).setEnabled(true); // reenables quit button
    }
    else if(event.getActionCommand().equals("Start")) {
      letters.dispose();
      this.dispose();
      if (bDEBUG) {
	for (int i = 0; i < input.length; i++) {
	  System.out.println("Input " + i + " = " + input[i]);
	}
      }
      ScrabbleFrame game = new ScrabbleFrame(input, gameSize[0], gameSize[1]);
      game.setVisible(true);
      game.validate();
    }
    this.setInput();
    this.validate();
  }

  /**
   * Draw random letters for the players and determines turn by who drew the 
   * lowest alphabetically.  Blanks automatically win and the rest go in rank 
   * with A being the best and Z the worst.
   */
  private int drawTurn() {
    int turn = -1;
    int debug = 0;
    boolean done = false;
    while (!done) {
      if (bDEBUG){
      debug = debug + 1;
      System.out.println("Loop " + debug + 
			 " until one person has highest letter");
      }
      turn = -1;
      if (letters != null)
	letters.dispose();
      Vector drawPile = new Vector();
      LetterModel[] drawnLetters = new LetterModel[numPlayers];
      int[] priorities = new int[numPlayers]; 
      ScrabbleFileReader tempLetters = new ScrabbleFileReader();
      int[] ranks = new int[numPlayers];
      tempLetters.readFile();
      drawPile = tempLetters.getLetters();
      for (int i = 0; i < drawnLetters.length;i++) {
	int letter = (int)(Math.random() * drawPile.size());
	drawnLetters[i] = (LetterModel)(drawPile.remove(letter));
	//priority detemined by the distance the letter is from A
	//the smallest distance alphabetically from A is first
	String curLetter = String.valueOf(drawnLetters[i].getLetter());
	priorities[i] = curLetter.compareTo("A");
	if(curLetter.equals("b")) // if you draw a blank you get the lowest score
	  priorities[i] = -1;
      }
      this.createDrawnLettersWindow(drawnLetters);

      for (int i = 0; i < numPlayers; i++) {
	for (int j = 0; j < numPlayers; j++) {
	  if (i != j && priorities[i] <= priorities[j])
	    ranks[i] = ranks[i] + 1;
	}  
      }
      for (int i = 0; i < numPlayers; i++) {
	if(bDEBUG)
	System.out.println("Turn = " + turn + input[i+5] + 
			   " rank = " + ranks[i]);	  

	if (ranks[i] == (numPlayers - 1)) {
	  if (turn == -1)
	    done = true;
	  else
	    done = false;
	  turn = i;
	}
      }
    }
    JOptionPane.showMessageDialog(letters, input[turn+5]  + " Goes first");
    return turn;
  }

  /**
   * Displays the letters drawn.
   */
  private void createDrawnLettersWindow(LetterModel[] drawnLetters) {
    letters = new JWindow(this);
    letters.setSize(300, 200);
    JPanel top = new JPanel();
    top.setLayout(new GridLayout(4, 2));
    JLabel[] namePanel = new JLabel[numPlayers];
    for (int i = 0; i < numPlayers; i++) {
      namePanel[i] = new JLabel(input[i+5] + " Drew a");
    }
    PlayerLetterButton[] letter = new PlayerLetterButton[numPlayers];
    for (int i = 0; i < numPlayers; i++) {
      letter[i] = new PlayerLetterButton(drawnLetters[i]);
    }
    for (int i = 0; i < numPlayers; i++) {
      top.add(namePanel[i]);
      top.add(letter[i]);
    }
    letters.getContentPane().add(top);
    letters.setVisible(true);
    letters.validate();
  }

  /**
   * Initializes all frames and sets up this <code>ScrabbleInit</code> frame.
   */
  private void addComponents() {
    // sets up the left panel of the display which holds start word lenght and
    // players names
    JPanel[] choiceAreas = setChoiceAreas();
    JPanel left = new JPanel();
    left.setLayout(new GridLayout(3, 1));
    left.add(choiceAreas[4]);
    left.add(choiceAreas[5]);
    left.add(choiceAreas[6]);
    this.getContentPane().add(left);

    // the panel for the right side
    JPanel right = new JPanel();
    right.setLayout(new GridLayout(2,1));

    // the top right panel
    JPanel rightTop = new JPanel();
    rightTop.setLayout(new GridLayout(2,2));
    for (int i = 0; i < choiceAreas.length - 3; i++) {
      rightTop.add(choiceAreas[i]);
    }
    right.add(rightTop);

    // my sig for this program (bottom right panel)
    MySigPanel rightBottom = new MySigPanel("Scrabble Version 1.3");
    right.add(rightBottom);

    this.getContentPane().add(right);
  }


  /**
   * Initializes all inner panels and radio button groups and other widgets.
   *
   * @return the 6 panels of this frame in a <code>JPanel[]</code>
   */
  private JPanel[] setChoiceAreas() {
    // intitalizes choice and button panels
    for (int i = 0; i < choicePanels.length; i++) {
      choicePanels[i] = new JPanel();
    }

    // the choose number of player panel and radio buttons
    ButtonGroup numPlayer = new ButtonGroup();
    for (int i = 0; i < players.length; i++) {
      players[i] = new JRadioButton(String.valueOf(i + 2));
      players[i].addActionListener(this);
      numPlayer.add(players[i]);
      choicePanels[0].add(players[i]);
    }
    players[0].setSelected(true); // 2 players set as default button
    choicePanels[0].setBorder(new TitledBorder(new BevelBorder(1),
					       "Number of Players"));

    // the bingo variation panel and radio buttons
    ButtonGroup bingo = new ButtonGroup();
    bingoButtons[0] = new JRadioButton("Yes");
    bingoButtons[0].addActionListener(this);
    bingo.add(bingoButtons[0]);
    choicePanels[1].add(bingoButtons[0]);
    bingoButtons[1] = new JRadioButton("No");
    bingoButtons[1].addActionListener(this);
    bingo.add(bingoButtons[1]);
    choicePanels[1].add(bingoButtons[1]);
    bingoButtons[1].setSelected(true); // bingo off set as default
    choicePanels[1].setBorder(new TitledBorder(new BevelBorder(1),
					       "Bingo Variation?"));

    // the replace blanks panel and radio buttons
    ButtonGroup blanks = new ButtonGroup();
    replaceBlanks[0] = new JRadioButton("Yes");
    replaceBlanks[0].addActionListener(this);
    blanks.add(replaceBlanks[0]);
    choicePanels[2].add(replaceBlanks[0]);
    replaceBlanks[1] = new JRadioButton("No");
    replaceBlanks[1].addActionListener(this);
    blanks.add(replaceBlanks[1]);
    choicePanels[2].add(replaceBlanks[1]);
    replaceBlanks[1].setSelected(true); // no replace blanks set as default
    choicePanels[2].setBorder(new TitledBorder(new BevelBorder(1),
					       "Replace Blanks?"));

    // the start game and quit buttons and panel for them
    JButton done = new JButton("Done");
    done.setActionCommand("Done");
    done.addActionListener(this);
    JButton quit = new JButton("Quit");
    quit.addActionListener(this);
    quit.setActionCommand("Quit");
    choicePanels[3].add(done);
    choicePanels[3].add(quit);

    // the start word length panel and radio buttons
    ButtonGroup startLength = new ButtonGroup();
    for (int i = 0; i < wordLengths.length; i++) {
      wordLengths[i] = new JRadioButton(String.valueOf(i + 2));
      wordLengths[i].addActionListener(this);
      startLength.add(wordLengths[i]);
      choicePanels[4].add(wordLengths[i]);
    }
    wordLengths[0].setSelected(true); // length of 2 set as default
    choicePanels[4].setBorder(new TitledBorder(new BevelBorder(1),
					       "Start Word Length"));

    // the names text fields and panel for them
    for (int i = 0; i < numPlayers; i++) {
      names[i] = new JTextField("Player " + (i + 1), 15);
      names[i].addActionListener(this);
      names[i].setActionCommand("Player " + (i + 1));
      choicePanels[5].add(names[i]);
    }
    choicePanels[5].setBorder(new TitledBorder(new BevelBorder(1),
					       "Player Names"));

    ButtonGroup screen = new ButtonGroup();
    for (int i = 0; i < screenSize.length; i++) {
      screenSize[i] = new JRadioButton(SCREEN_SIZE[i]);
      screenSize[i].addActionListener(this);
      screen.add(screenSize[i]);
      choicePanels[6].add(screenSize[i]);
    }
    screenSize[0].setSelected(true); // 640x480 set as default
    choicePanels[6].setBorder(new TitledBorder(new BevelBorder(1),
					       "Screen Size"));
    // returns the creates choice button panels
    return choicePanels;
  }

  /**
   * When a choice is changed or the start game button has been pressed this
   * is called to set all of the <code>input[]</code> entries to the changed
   * values.
   */
  private void setInput() {
    // if there are only two players this sets that input and makes sure there
    // are only two text fields for names
    if(players[0].isSelected()) {
      if (input[0].equals("3") || input[0].equals("4")) {  
	choicePanels[5].remove(names[2]);
	names[2] = null;
	choicePanels[5].repaint();
      }
      if (input[0].equals("4")) {
	choicePanels[5].remove(names[3]);
	names[3] = null;
	choicePanels[5].repaint();
      }
      input[0] = "2";
    }
    // if there are only 3 players this sets that input and makes sure there
    // are only three text fields for names
    else if(players[1].isSelected()) {
      if (input[0].equals("4")) {  
	choicePanels[5].remove(names[3]);
	names[3] = null;
	choicePanels[5].repaint();
      }
      else if (input[0].equals("2")) {
	names[2] = new JTextField("Player 3", 15);
	names[2].addActionListener(this);
	names[2].setActionCommand("Player 3");
	choicePanels[5].add(names[2]);
      }
      input[0] = "3";
    }
    // if there are only 4 players this sets that input and makes sure there
    // are only four text fields for names
    else if(players[2].isSelected()) {
      if (input[0].equals("2")) {
	names[2] = new JTextField("Player 3", 15);
	names[2].addActionListener(this);
	names[2].setActionCommand("Player 3");
	choicePanels[5].add(names[2]);
      }
      if (input[0].equals("2") || input[0].equals("3")) {  
	names[3] = new JTextField("Player 4", 15);
	names[3].addActionListener(this);
	names[3].setActionCommand("Player 4");
	choicePanels[5].add(names[3]);
      }
      input[0] = "4";
    }
    //if block for setting word length
    if(wordLengths[0].isSelected())
      input[1] = "2";
    else if(wordLengths[1].isSelected())
      input[1] = "3";
    else if(wordLengths[2].isSelected())
      input[1] = "4";
    else if(wordLengths[3].isSelected())
      input[1] = "5";
    else if(wordLengths[4].isSelected())
      input[1] = "6";
    else if(wordLengths[5].isSelected())
      input[1] = "7";
    
    // if block for setting the bingo variation
    if(bingoButtons[0].isSelected())
      input[3] = "true";
    else if(bingoButtons[1].isSelected())
      input[3] = "false";
    
    // if block for setting the replace blank variation
    if(replaceBlanks[0].isSelected())
      input[4] = "true";
    else if(replaceBlanks[1].isSelected())
      input[4] = "false";

    for (int i = 0; i < screenSize.length; i++) {
      if(screenSize[i].isSelected()) {
	StringTokenizer sizeToken = new StringTokenizer(SCREEN_SIZE[i], "x");
	String[] size = new String[2];
	int index = 0;
	while (sizeToken.hasMoreTokens()) {
	  size[index] = sizeToken.nextToken();
	  gameSize[index] = Integer.parseInt(size[index]);
	  index++;
	}
      }
    }
    
    // sets names from the text fields
    input[5] = names[0].getText();
    input[6] = names[1].getText();
    if (names[2] != null)
      input[7] = names[2].getText();
    if (names[3] != null)
      input[8] = names[3].getText();

    if (bDEBUG) {
      for (int i = 0; i < input.length; i++) {
	System.out.println("Input " + i + " = " + input[i]);
      }
    }

    numPlayers = Integer.parseInt(input[0]);
  }

  /**
   * My thanks for playing frame.
   */
  private void thanks() {
    JFrame thanks = new JFrame();
    thanks.setSize(300, 300);
    thanks.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    thanks.setVisible(true);
    thanks.setTitle("Thanks for Playing");
    thanks.getContentPane().add(new MySigPanel("Thanks for playing.  Coded by"));
    thanks.validate();
  }

}// end of class ScrabbleInit
