import java.util.Vector;

/**
 * CS1322: Programming Assignment #5 - Spring 2003 - Phase 5
 *
 *  <PRE>
 *  Class ScrabbleModel.java- The "main" class for Scrabble.  Does all of the
 *  calculations and such in an interface independent way so that one could
 *  play the game using a Graphical or text-based interface.
 *
 *  Revisions: 1.0 Apr. 06, 2003
 *               Created the ScrabbleModel class
 *
 *  </PRE>
 *
 *  Collaboration statement:
 *  "I worked on the homework assignment alone, using only course materials."
 *
 *  @author <A HREF="mailto:gtg445h@mail.gatech.edu">Chris Wells</A>
 *  @version Version 1.0, Apr. 06, 2003
 */

public class ScrabbleModel implements Scrabble, ScrabbleConstants{

  /**
   * A constant to trigger debug statements
   */

  public static final boolean bDEBUG = false;

  /**
   * Holds the representations of the <code>board</code> squares.
   *
   */
  private SquareModel[][] board;

  /**
   * The number of players in the game.  A number from 2-4.
   */
  private int numPlayers;

  /**
   * An array to hold the players.
   */
  private PlayerModel[] players;

  /**
   * A <code>Vector</code> to hold the letters in the "bag"
   */
  private Vector lettersRemaining;

  /**
   * The letter the player is currently considering putting down.
   */
  private LetterModel current;

  /**
   * The letters that have already been played in the squares they have
   * been placed in.
   */
  private Vector lettersPlaced;

  /**
   * Holds the number of the player whos turn it is.
   */
  private int turn;

  /**
   * Determins weathere the first move has been made or not.
   * True if its the first move, false otherwise.
   */
  private boolean start;

  /**
   * Determines if the end game condition has been met or not.
   */
  private boolean loseCondition;

  /**
   * True if the variation to do Bingo rules has been chosen, false otherwise
   * (A Bingo is when all 7 letter on the players tray are played in one turn.)
   */
  private boolean bingo;

  /**
   * True if the variation to replace blanks on the board with the same
   * letter on their tray has been chosen, false otherwise.
   */
  private boolean replaceBlanks;

  /**
   * Determines weather a player is replacing a blank piece or not.
   */
  private boolean replaceBlank;

  /**
   * Determines weather a player has requested to exchange letters or not
   */
  private boolean replaceLetters;

  /**
   * Counts the number of turns the game has gone without a play
   */
  private int numOfTurnsWithoutPlay;

  /**
   * Holds the length the beginning word should be at least as long as.
   * Should be 2-7 but more than likely somewhere between 2-4.
   */
  private int beginWordLength;

  /**
   * Creates a new <code>ScrabbleModel</code> instance. It turns off all
   * variations by defaule and sets word length to 0 and sets number of players
   * to 4.
   *
   */
  public ScrabbleModel() {
    this(false, false, 2, 4, 0);
  }

  /**
   * Creates a new <code>ScrabbleModel</code> instance.
   *
   * @param bingo true if the bingo variation has been chosen
   * @param replaceBlanks true if the replace blanks variation has been chosen
   * @param beginWordLength the beginning word length (1-7)
   * @param numPlayers the number of players (2-4)
   * @param turn the starting player number cause he goes first
   */
  public ScrabbleModel(boolean bingo, boolean replaceBlanks,
		       int beginWordLength, int numPlayers,
		       int turn) {
    this.bingo = bingo;
    this.replaceBlanks = replaceBlanks;
    this.replaceBlank = false;
    this.replaceLetters = false;
    this.numOfTurnsWithoutPlay = 0;
    this.beginWordLength = beginWordLength;
    this.numPlayers = numPlayers;
    this.turn = turn;
    // each interface will determine how to draw for the turn
    this.players = new PlayerModel[this.numPlayers];
    for(int i = 0; i < numPlayers; i++) {
      players[i] = new PlayerModel();
    }
    this.current = null;
    this.start = true;
    this.loseCondition = false;
    this.lettersPlaced = new Vector();
    this.readDataFile();
    if (bDEBUG)
      System.out.println("Bingo = " + bingo + " Blanks = " + replaceBlanks
			 + " Begin Word Length = " + beginWordLength
			 + " Number of Players" +numPlayers + '\n' +
			 " Player array size = " + players.length + 
			 " Number of letters in bag" + lettersRemaining.size() +
			 " turn = " + turn + " start = " + start);
    
    for (int i = 0; i < numPlayers; i++) {
      for (int j = 0; j < TRAY_SIZE; j++) {
	int letter = (int)(Math.random() * lettersRemaining.size());
	if(bDEBUG)
	  System.out.println("Letter var = " + letter + " i = " + i + " j = " + 
			     j);
	
	players[i].addLetter((LetterModel)lettersRemaining.get(letter));
	lettersRemaining.remove(letter);
      }
    }
  }

  /**
   * Reads the scrabble data file in and initializes <code>board</code>
   * and <code>lettersRemaining</code>.
   *
   */
  private void readDataFile() {
    ScrabbleFileReader gameInfo = new ScrabbleFileReader();
    gameInfo.readFile();
    this.lettersRemaining = gameInfo.getLetters();
    this.board = gameInfo.getBoard();

  }

  /**
   * Returns the ith player. The parameter is zero indexed.
   *
   * @param i the index of the player to retrieve
   * @return the PlayerModel at location i
   */
  public PlayerModel getPlayer(int i){
    return players[i];
  }

  /**
   * Returns the current score of Player i.
   *
   * @param i the zero based index of the Player
   * @return the score of Player i
   */
  public int getScore(int i) {
    return (players[i]).getScore();
  }

  /**
   * Returns a Vector of Player i's letters in his/her tray. 
   * The letters returned are LetterModels. 
   *
   * @param i the zero based index of the Player
   * @return the Vector of Player i's LetterModels in their tray.
   */
  public Vector getPlayerLetters(int i) {
    return (players[i]).getLetters();
  }

  /**
   * Returns the index of the Player who's curently at play (who's turn it is). 
   *
   * @return the zero based index of the Player who's turn it is.
   */
  public int getTurn() {
    return turn;
  }

  /**
   * Returns whether or not a game ending condition exists.
   *
   * @return whether or not a game ending condition exists
   */
  public boolean gameIsOver() {
    if(loseCondition)
      return true;
    else
      return false;
  }

  /**
   * returns all SquareModels in array form.
   *
   * @return the array of SquareModels
   */
  public SquareModel[][] getSquares() {
    return board;
  }

  /**
   * Returns the SquareModel on the board at ROW i, COLUMN j. The indecies are
   * zero based. 
   *
   * @param i the ROW of the SquareModel to return
   * @param j the COLUMN of the SquareModel to return
   * @return the SquareModel on the board at ROW i, COLUMN j.
   */
  public SquareModel getSquare(int i, int j) {
    return board[i][j];
  }

  /**
   * Sets the currently selected letter of the game to the LetterModel passed
   * in. This method does not lay down a letter onto the board. It simply 
   * "picks up" a letter tile and awaits instructions to place it at a location.
   *
   * @param current the letter to "pick up"
   */
  public void setCurrent(LetterModel current) {
    if (current.getLetter() == '#' && replaceBlanks == true) {
      replaceBlank = true;
      if (bDEBUG)
	System.out.println("Setting replaceBlank to true");

      return;
    }
    else if (replaceBlank == true) {
      replaceBlank(current);
      replaceBlank = false;
      return;
    }
    else if (current.getLetter() == '^' && (lettersRemaining.size() != 0)) {
      replaceLetters = true;
      if (bDEBUG)
	System.out.println("Setting replaceLetter to true");

      return;
    }
    else if (replaceLetters == true) {
      lettersPlaced.add(current);
      PlayerModel currentPl = getPlayer(getTurn());
      currentPl.removeLetter(current);
      return;
    }
    else if (current.getLetter() == '!' && lettersPlaced.size() != 0) {
      lettersPlaced.remove(lettersPlaced.size() - 1);
      return;
    }
    else if (current.getLetter() == '^' || current.getLetter() == '#') {
      return;
    }
    this.current = current;
    if (bDEBUG)
      System.out.println("Current letter is " + current.getLetter());
    
  }

  /**
   * Replaces a blank on the board with an appropriate letter from the player.
   *
   * @param current the letter to replace the blank
   */  
  private void replaceBlank(LetterModel current) {
    PlayerModel currentPl = getPlayer(getTurn());
    for (int i = 0; i < Y_LENGTH; i++) {
      for (int j = 0; j < X_LENGTH; j++) {
	SquareModel temp = getSquare(i, j);
	LetterModel tempL = temp.getLetter();
	if (tempL != null && tempL.getValue() == 0
	    && (current.getLetter() == tempL.getLetter()))
	  {
	    if (bDEBUG)
	      System.out.println("Replacing blank ...");

	    temp.setLetter(current);
	    currentPl.removeLetter(current);
	    currentPl.addLetter(new LetterModel('b',0));
	    return;
	  }
      }
    }
  }

  /**
   * Places the current LetterModel onto the SquareModel passed in thus placing 
   * the letter onto the board. This method should clear the current letter as 
   * set by setCurrent(LetterModel). This method should do nothing if:
   *
   * The SquareModel passed in already contains a letter
   * No tile has been placed yet and the SquareModel is not the START square
   * the current letter is null, i.e. setCurrent(LetterModel) has not yet been 
   * called. 
   *
   * @param square the SquareModel on which to place the current letter
   */
  public void squareSelect(SquareModel square) {
    if(square.getLetter() == null && current != null && 
       (start == true && square.getSquareType() == START)) {
      square.setLetter(current);
      lettersPlaced.add(square);
      getPlayer(getTurn()).removeLetter(current);
      start = false;
    if (bDEBUG)
      System.out.println("Current letter is " + current.getLetter() +
			 " set start to false" + '\n' + "Square is now(" +
			 square.getXCoord() + ", " + square.getYCoord() + ", " +
			  square.getLetter().getLetter() + ")");
    }
    else if(square.getLetter() == null && current != null && 
	    start == false) {
      square.setLetter(current);
      lettersPlaced.add(square);
      getPlayer(getTurn()).removeLetter(current);
    if (bDEBUG)
      System.out.println("Current letter is " + current.getLetter() +
			 " set start to false" + '\n' + "Square is now(" +
			 square.getXCoord() + ", " + square.getYCoord() + ", " +
			  square.getLetter().getLetter() + ")");
    }
  }

  /**
   * A var to store weather the <code>getPlayScore</code> method should
   * bother to keep looking left to right if true or up and down if false.
   * This is because the <code>lettersPlaced</code> should either be horiz
   * or vert and therefore in the same word.
   */

  private boolean leftRight;

  /**
   * Determines the score for this play and adds it to the players score.
   */
  private int getPlayScore() {
    int score = 0;
    PlayerModel currentPl = getPlayer(getTurn());
    leftRight = false; 
    // determines weather the current letters are going left to right
    // or up and down.  Reduces having to look at all for all the letters
    // left and right since they should be in horizontal row (true) and therefore
    // part of the same new word formed.  If it is false that means the
    // letters currently placed are in a verticle row so there is no need
    // to check anyMore than the first letter up and down since the rest
    // would be in the same word
    if (lettersPlaced.size() == 0)
      return 0;
    for (int i = 0; i < lettersPlaced.size(); i++) {
      score = score + findWordLeftRight((SquareModel)(lettersPlaced.get(i)));
      // no need to keep looking so exit the loop
      if (leftRight)
	i = lettersPlaced.size();
      
    }
    for (int i = 0; i < lettersPlaced.size(); i++) {
      score = score + findWordUpDown((SquareModel)(lettersPlaced.get(i)));
      // no need to keep looking so exit the loop
      if (!(leftRight))
	i = lettersPlaced.size();
    }
    if (lettersPlaced.size() == TRAY_SIZE && bingo == true)
      score = score + 50;
    return score;
  }

  /**
   * Finds all letters to the left and right in the current word.
   *
   * @param square a <code>SquareModel</code> which holds the letter
   * to look around.
   */
  private int findWordLeftRight(SquareModel square) {
    int xCoord = square.getXCoord();
    int yCoord = square.getYCoord();
    int wordScore = 0;
    if ((xCoord == (X_LENGTH - 1) ||
	 getSquare(yCoord, xCoord + 1).getLetter() == null)
	 &&
	(xCoord == 0 || getSquare(yCoord, xCoord - 1).getLetter() == null)) { 
      // both sides are empty meaning it didnt make a new word
      return 0;
    }
    Vector word = new Vector(); // the current word formed
    int wordMult = 1; // muliplier for word
    int letterMult = 1; // multiplier for individual letter
    // add current letter to word
    word.add(getSquare(yCoord, xCoord));
    // find right side of the word
    xCoord++;
    while (xCoord < X_LENGTH && getSquare(yCoord, xCoord).getLetter() != null) {
      if (bDEBUG)
	System.out.println("xCoord is" + xCoord);

      word.add(getSquare(yCoord, xCoord));
      xCoord++;
    }
    // find left side of the word
    xCoord = square.getXCoord() - 1;
    while (xCoord >= 0 && getSquare(yCoord, xCoord).getLetter() != null) {
      if (bDEBUG)
	System.out.println("xCoord is" + xCoord);
      
      word.add(getSquare(yCoord, xCoord));
      xCoord--;
    }
    int type = 0; // type of square
    SquareModel current;
    for (int i = 0; i < word.size(); i++) {
      current = (SquareModel)(word.get(i));
      if(i != 0 && lettersPlaced.contains(word.get(i))) {
	// dont care if the current letter is in the lettersPlaced we already
	// know that
	leftRight = true;
      }
      if(lettersPlaced.contains(word.get(i))) {
	// know the letter has been played this turn
	type = current.getSquareType();
	if (type == START || type == DOUBLE_WORD)
	  wordMult = wordMult * 2;
	else if (type == TRIPLE_WORD)
	  wordMult = wordMult * 3;
	else if (type == TRIPLE_LETTER)
	  letterMult = letterMult * 3;
	else if (type == DOUBLE_LETTER)
	  letterMult = letterMult * 2;
      }
      wordScore = wordScore + (current.getLetter()).getValue() * letterMult;
      letterMult = 1; //reset for next letter
    }
    wordScore = wordScore * wordMult;
    return wordScore;
  }

  /**
   * Finds all letters to the top and bottom in the current word.
   *
   * @param square a <code>SquareModel</code> which holds the letter
   * to look around.
   */
  private int findWordUpDown(SquareModel square) {
    int xCoord = square.getXCoord();
    int yCoord = square.getYCoord();
    int wordScore = 0;
    if ((yCoord == (Y_LENGTH - 1) ||
	 getSquare(yCoord + 1, xCoord).getLetter() == null) &&
	(yCoord == 0 || getSquare(yCoord - 1, xCoord).getLetter() == null)) { 
      // both sides are empty meaning it didnt make a new word
      return 0;
    }
    Vector word = new Vector(); // the current word formed
    int wordMult = 1; // muliplier for word
    int letterMult = 1; // multiplier for individual letter
    // add current letter to word
    word.add(getSquare(yCoord, xCoord));
    // find bottom of the word
    yCoord++;

    while (yCoord < Y_LENGTH && getSquare(yCoord, xCoord).getLetter() != null) {
      if (bDEBUG)
	System.out.println("yCoord is" + xCoord);

      word.add((getSquare(yCoord, xCoord)));
      yCoord++;
    }
    // find top of the word
    yCoord = square.getYCoord() - 1;
    while (yCoord >= 0 && getSquare(yCoord, xCoord).getLetter() != null) {
      if (bDEBUG)
	System.out.println("yCoord is" + xCoord);

      word.add(getSquare(yCoord, xCoord));
      yCoord--;
    }
    int type = 0; // type of square
    SquareModel current;
    for (int i = 0; i < word.size(); i++) {
      current = (SquareModel)(word.get(i));
      if(i != 0 && lettersPlaced.contains(word.get(i))) {
	// dont care if the current letter is in the lettersPlaced we already
	// know that
	leftRight = false;
      }
      if(lettersPlaced.contains(word.get(i))) {
	// know the letter has been played this turn
	type = current.getSquareType();
	if (type == START || type == DOUBLE_WORD)
	  wordMult = wordMult * 2;
	else if (type == TRIPLE_WORD)
	  wordMult = wordMult * 3;
	else if (type == TRIPLE_LETTER)
	  letterMult = letterMult * 3;
	else if (type == DOUBLE_LETTER)
	  letterMult = letterMult * 2;
      }
      wordScore = wordScore + (current.getLetter()).getValue() * letterMult;
      letterMult = 1; //reset for next letter
    }
    wordScore = wordScore * wordMult;
    return wordScore;
  }

  /**
   * Should be called to end a player's turn, calculate his/her score for the
   * letters placed on this turn and proceed to the next player.
   *
   */
  public void endTurn() {
    PlayerModel currentPl = getPlayer(getTurn());
    if (replaceLetters) {
      if (lettersPlaced.size() > lettersRemaining.size()) {
	current = null;
	for (int i = 0; i < lettersPlaced.size();i++ ) {
	  // not enough letters so the letters the player put in are given
	  // back
	  currentPl.addLetter((LetterModel)lettersPlaced.get(i));
	}
	replaceLetters = false;
	lettersPlaced = new Vector();
	return;
      }
      for (int i = 0; i < lettersPlaced.size(); i++) {
	int letter = (int)(Math.random() * lettersRemaining.size());
	currentPl.addLetter((LetterModel)lettersRemaining.get(letter));
	lettersRemaining.remove(letter);
      }
      lettersRemaining.addAll(lettersPlaced);
      current = null;
      replaceLetters = false;
      lettersPlaced = new Vector();
      if (getTurn() == (numPlayers - 1)) {
	turn = 0;
      }
      else {
	turn = turn + 1;
      }
      numOfTurnsWithoutPlay++; 
      return;
    }
    currentPl.incrementScore(getPlayScore());

    if (lettersRemaining.size() != 0) {
      for (int i = (getPlayerLetters(getTurn())).size(); i < TRAY_SIZE;i++) {
	if (lettersRemaining.size() != 0) {
	  int letter = (int)(Math.random() * lettersRemaining.size());
	  currentPl.addLetter((LetterModel)lettersRemaining.get(letter));
	  lettersRemaining.remove(letter);
	}
      }
    }
    else if(getPlayerLetters(getTurn()).size() == 0) {
      // one player "won" went out of letters
      setFinalScores(true);
      loseCondition = true;
      return;
    }
    if (getPlayScore() == 0) {
      numOfTurnsWithoutPlay++; 
    }
    else if(getPlayScore() != 0) {
      numOfTurnsWithoutPlay = 0; 
    }
    if (numOfTurnsWithoutPlay == numPlayers) {
      setFinalScores(false);
      loseCondition = true;
      return;
    }
    if (bDEBUG)
      System.out.println(" Score for turn = " + getPlayScore() + '\n' +
			 "players score is " + getScore(turn));
    if (getTurn() == (numPlayers - 1)) {
      turn = 0;
    }
    else {
      turn = turn + 1;
    }

    if (bDEBUG) {
      System.out.println("End players turn -- turn = " + turn);
    }

    current = null;
    lettersPlaced = new Vector();
  }

  /**
   * Sets the final scores as set forth in the Scrabble manual.
   *
   * @param someoneWon a boolean true if someone has no letters and the
   * game ends that way, false otherwise.
   */
  private void setFinalScores(boolean someoneWon) {
    if (someoneWon) {
      int scores = 0;
      int subScore = 0;
      for (int i = 0; i < numPlayers; i++) {
	if (i != getTurn()) { // not the winning player
	  subScore = 0; // score to subtract
	  // temp var to store player letters
	  Vector letter = (getPlayerLetters(i)); 
	  while (letter.size() > 0) {
	    LetterModel currentLetter = (LetterModel)letter.get(0);
	    (getPlayer(i)).removeLetter(currentLetter);
	    letter = (getPlayerLetters(i));
	    subScore = subScore + currentLetter.getValue();
	  }
	  (getPlayer(i)).incrementScore(-subScore);
	  if (bDEBUG) {
	    System.out.println("Subtracted " + subScore +
			       " points and current score is " + 
			       getScore(i));
	  }
	  scores = scores + subScore;
	}
      }
      getPlayer(getTurn()).incrementScore(scores); // winning player gets the
      // points subtracted from the
      //losing players
      if (bDEBUG) {
	System.out.println("Added " + scores +
			   " points and current score is " + 
			   getScore(getTurn()));
      }
      return;
    }
    int subScore = 0;
    for (int i = 0; i < numPlayers; i++) {
      subScore = 0;    
      Vector letter = (getPlayerLetters(i)); // temp var to store player letters
      while ((getPlayerLetters(i)).size() > 0) {
	LetterModel currentLetter = (LetterModel)letter.get(0);
	(getPlayer(i)).removeLetter(currentLetter);
	subScore = subScore + currentLetter.getValue();
      }
      (getPlayer(i)).incrementScore(-subScore);
      if (bDEBUG) {
	System.out.println("Subtracted " + subScore +
			   " points and current score is " + 
			   getScore(i));
      }
    }
  }

}// end of class ScrabbleModel
