import javax.swing.JButton;
import java.awt.Graphics;
import java.awt.Graphics2D;
import javax.swing.ImageIcon;
import java.awt.geom.Rectangle2D;
import java.awt.font.FontRenderContext;
import java.awt.Font;
import java.awt.Color;

/**
 * CS1322: Programming Assignment #6 - Spring 2003 - Phase 3
 *
 *  <PRE>
 *  Class SquareButton.java- A button to represent a single square in Scrabble.
 *  Used to make the user interface easier.
 *
 *  Revisions: 1.0 Apr. 17, 2003
 *               Created the SquareButton class
 *
 *  </PRE>
 *
 *  Collaboration statement:
 *  "I worked on the homework assignment alone, using only course materials."
 *
 *  @author <A HREF="mailto:gtg445h@mail.gatech.edu">Chris Wells</A>
 *  @version Version 1.0, Apr. 17, 2003
 */

public class SquareButton extends JButton implements ScrabbleConstants {

  /**
   * A constant to trigger debug events.
   */
  public final boolean bDEBUG = false;

  /**
   * The square this button represents.
   */
  private SquareModel square;

  /**
     A conversion factor that determines where to place the value of
     letter relative to the position of the letter and scaled by the
     size of the button in the x direction.
  */
  private double valuePlacementXConvFactor;

  /**
     A conversion factor that determines where to place the value of
     letter relative to the position of the letter and scaled by the
     size of the button in the y direction.
  */
  private double valuePlacementYConvFactor;

  /**
     The font size that looks the best on a 800 by 600 pixel frame.  Scaled to
     fit current letter button size.
  */
  private final int DEFAULT_FONT_SIZE = 31;

  /**
     The height of the letter on a 800 x 600 pixel frame
  */
  private final double DEFAULT_HEIGHT = 47.0;

  /**
   * Creates a new <code>SquareButton</code> by setting the square variable
   * to the representative variable.
   */
  public SquareButton(SquareModel square) {
    this.square = square;
    valuePlacementXConvFactor = this.getWidth()/5;
    valuePlacementYConvFactor = this.getHeight()/9;
  }

  /**
   * Overwrites the <code>paintComponent</code> of the button to display what 
   * kind of button it is and the letter that has been placed on it.
   */
  protected void paintComponent(Graphics g) {

    Graphics2D g2 = (Graphics2D) g;
    int type = 0;
    int squareWidth = this.getWidth();
    int squareHeight = this.getHeight();
    type = square.getSquareType();
    String pic = "";
    if (type == TRIPLE_LETTER) { 
      pic = "TripleLetter.jpg";
      this.setToolTipText("A Triple Letter Square.  Click to place letter");
    }
    else if(type == DOUBLE_LETTER) {
      pic = "DoubleLetter.jpg";
      this.setToolTipText("A Double Letter Square.  Click to place letter");
    }
    else if(type == DOUBLE_WORD) {
      pic = "DoubleWord.jpg";
      this.setToolTipText("A Double Word Square.  Click to place letter");
    }
    else if(type == TRIPLE_WORD) {
      pic = "TripleWord.jpg";
      this.setToolTipText("A Triple Word Square.  Click to place letter");
    }
    else if(type == START) {
      pic = "StartSquare.jpg";
      this.setToolTipText("The Start Square.  Click to place letter");
    }
    else {
      pic = "RegularSquare.jpg";
      this.setToolTipText("A Regular Square.  Click to place letter");
    }
    ImageIcon squareImage = new ImageIcon(pic);
    g.drawImage(squareImage.getImage(),0,0,squareWidth,
		squareHeight,this);
    if (square.getLetter() == null) {
      this.setEnabled(true);
    }
    else {
      this.drawLetter(g2);
      this.setEnabled(false);
    }
  }

  /**
   * A helper of the <code>paintComponent</code> method. Draw the letter
   * associated with this square if there is one.
   */
  private void drawLetter(Graphics2D g2) {
    ImageIcon letterBack = new ImageIcon("LetterTile.jpg");
    // the 2 is used so that the letter tile doesn't cover the whole square
    g2.drawImage(letterBack.getImage(),2,2,this.getWidth() - 2,
		 this.getHeight() - 2,this);
    // checks to see if the letter is not a blank.
    if (square.getLetter().getLetter() != 'b') {
      // font size of the letter. scaled by the height or width,
      // depenging on which one is smaller. of the button
      int letterSize = 0;
      if (this.getHeight() < this.getWidth()) { 
	letterSize = (int) (DEFAULT_FONT_SIZE * 
			    (this.getHeight() / DEFAULT_HEIGHT));
      }
      else {
	letterSize = (int) (DEFAULT_FONT_SIZE * 
			    (this.getWidth() / DEFAULT_HEIGHT));
      }
	
      if (bDEBUG)
	System.out.println("LetterFontSize = " + letterSize);
	
      Font letterFont = new Font("SansSerif", Font.PLAIN, letterSize);
      g2.setFont(letterFont);
      g2.setColor(Color.white);
      // This next sequence of variables is used to perfectly center
      // the letter at all times
      String buttonLetter = String.valueOf(square.getLetter().getLetter());
      FontRenderContext context = g2.getFontRenderContext();
      Rectangle2D bounds = letterFont.getStringBounds(buttonLetter,context);
      double lAscent = -bounds.getY();
      double lDescent = bounds.getHeight() - lAscent;
      double lHeight = bounds.getHeight();
      double lWidth = bounds.getWidth();
      double lBase = ((this.getHeight() - lHeight) / 2) + lAscent;
      double lLeft = (this.getWidth() - lWidth) / 2;
      // the factor of how big the letter size should be to the
      // letter set as 3:1
      int letterToValueFontConv = 3;
      g2.drawString(buttonLetter, (float)lLeft, (float)lBase);
      // the font size of the value
      int valueSize = (int) (letterSize /letterToValueFontConv);

      if (bDEBUG)
	System.out.println("ValueFontSize = " + valueSize);    

      Font valueFont = new Font("SansSerif", Font.PLAIN, valueSize);
      g2.setFont(valueFont);
      // places the value of the letter down and to the left of the letter
      String buttonValue = String.valueOf(square.getLetter().getValue());
      double vLeft = (double)(lLeft + lWidth + valuePlacementXConvFactor);
      double vBase = (double)(lBase + valuePlacementYConvFactor);
      g2.drawString(buttonValue, (float)vLeft, (float)vBase);
    }
  }

  /**
   * Returns the square model associated with this button.
   *
   * @return the square of this button.
   */
  public SquareModel getSquare() {
    return square;
  }
}// end of class SquareButton
